#include <u.h>
#include <libc.h>
#include <thread.h>
#include <ctype.h>
#include "acme.h"

static Ioproc *io;
Win *windows;
static Win *last;

void
winclosefiles(Win *w)
{
	if(w->ctl.io != nil){
		ioclose(w->ctl.io, w->ctl.fd);
		closeioproc(w->ctl.io);
		w->ctl.io = nil;
	}
	if(w->body.io != nil){
		ioclose(w->body.io, w->body.fd);
		closeioproc(w->body.io);
		w->body.io = nil;
	}
	if(w->addr.io != nil){
		ioclose(w->addr.io, w->addr.fd);
		closeioproc(w->addr.io);
		w->addr.io = nil;
	}
	if(w->tag.io != nil){
		ioclose(w->tag.io, w->tag.fd);
		closeioproc(w->tag.io);
		w->tag.io = nil;
	}
	if(w->event.io != nil){
		ioclose(w->event.io, w->event.fd);
		closeioproc(w->event.io);
		w->event.io = nil;
	}
	if(w->data.io != nil){
		ioclose(w->data.io, w->data.fd);
		closeioproc(w->data.io);
		w->data.io = nil;
	}
	if(w->xdata.io != nil){
		ioclose(w->xdata.io, w->xdata.fd);
		closeioproc(w->xdata.io);
		w->xdata.io = nil;
	}
}

void
winfree(Win *w)
{
	winclosefiles(w);
	if(w->c){
		chanfree(w->c);
		w->c = nil;
	}
	if(w->next)
		w->next->prev = w->prev;
	else
		last = w->prev;
	if(w->prev)
		w->prev->next = w->next;
	else
		windows = w->next;
	free(w);
}

void
windeleteall(void)
{
	Win *w, *next;

	for(w=windows; w; w=next){
		next = w->next;
		winctl(w, "delete");
	}
}

static Cfd *
wfid(Win *w, char *name)
{
	char buf[100];
	Cfd *fid = nil;

	if(strcmp(name, "ctl") == 0)
		fid = &(w->ctl);
	else if(strcmp(name, "body") == 0)
		fid = &(w->body);
	else if(strcmp(name, "addr") == 0)
		fid = &(w->addr);
	else if(strcmp(name, "tag") == 0)
		fid = &(w->tag);
	else if(strcmp(name, "event") == 0)
		fid = &(w->event);
	else if(strcmp(name, "data") == 0)
		fid = &(w->data);
	else if(strcmp(name, "xdata") == 0)
		fid = &(w->xdata);
	else{
		sysfatal("bad window file name %s", name);
	}

	if(fid->io == nil){
		snprint(buf, sizeof buf, "/mnt/acme/%d/%s", w->id, name);
		fid->io = ioproc();
		if(fid->io == nil)
			sysfatal("ioproc() failed: %r");
		fid->fd = ioopen(fid->io, buf, ORDWR);
		if(fid->fd < 0)
			sysfatal("cannot open %s: %r", buf);
	}
	return fid;
}

int
winopenfd(Win *w, char *name, int mode)
{
	char buf[100];
	
	snprint(buf, sizeof buf, "/mnt/acme/%d/%s", w->id, name);
	return ioopen(io, buf, mode);
}

int
winctl(Win *w, char *fmt, ...)
{
	char *s;
	va_list arg;
	Cfd *fid;
	int n;

	va_start(arg, fmt);
	s = evsmprint(fmt, arg);
	va_end(arg);

	fid = wfid(w, "ctl");
	n = iowrite(fid->io, fid->fd, s, strlen(s));
	free(s);
	return n;
}

int
winname(Win *w, char *fmt, ...)
{
	char *s;
	va_list arg;
	int n;

	va_start(arg, fmt);
	s = evsmprint(fmt, arg);
	va_end(arg);

	n = winctl(w, "name %s\n", s);
	free(s);
	return n;
}

int
winprint(Win *w, char *name, char *fmt, ...)
{
	char *s;
	va_list arg;
	int n;
	Cfd *fid;

	va_start(arg, fmt);
	s = evsmprint(fmt, arg);
	va_end(arg);

	fid = wfid(w, name);
	n = iowrite(fid->io, fid->fd, s, strlen(s));
	free(s);
	return n;
}

int
winaddr(Win *w, char *fmt, ...)
{
	char *s;
	va_list arg;
	int n;
	Cfd *fid;

	va_start(arg, fmt);
	s = evsmprint(fmt, arg);
	va_end(arg);
	
	fid = wfid(w, "addr");
	n = iowrite(fid->io, fid->fd, s, strlen(s));
	free(s);
	return n;
}

int
winreadaddr(Win *w, uint *q1)
{
	char buf[40], *p;
	uint q0;
	int n;
	Cfd *fid;

	fid = wfid(w, "addr");
	n = ioread(fid->io, fid->fd, buf, sizeof buf-1);
	if(n <= 0)
		return -1;
	buf[n] = 0;
	q0 = strtoul(buf, &p, 10);
	if(q1)
		*q1 = strtoul(p, nil, 10);
	return q0;
}

int
winread(Win *w, char *file, void *a, int n)
{
	Cfd *fid = wfid(w, file);

	return ioread(fid->io, fid->fd, a, n);
}

int
winwrite(Win *w, char *file, void *a, int n)
{
	Cfd *fid = wfid(w, file);

	return iowrite(fid->io, fid->fd, a, n);
}

char*
winmread(Win *w, char *file)
{
	char *buf;
	int n, tot, m;
	Cfd *fid;
	
	m = 128;
	buf = emalloc(m+1);
	tot = 0;
	fid = wfid(w, file);
	while((n = ioread(fid->io, fid->fd, buf+tot, m-tot)) > 0){
		tot += n;
		if(tot >= m){
			m += 128;
			buf = erealloc(buf, m+1);
		}
	}
	if(n < 0){
		free(buf);
		return nil;
	}
	buf[tot] = 0;
	return buf;
}

static long
_ioseek(va_list *arg)
{
	int fd;
	vlong n;
	int off;

	fd = va_arg(*arg, int);
	n = va_arg(*arg, vlong);
	off = va_arg(*arg, int);
	return seek(fd, n, off);
}

int
winseek(Win *w, char *file, int n, int off)
{
	Cfd *fid = wfid(w, file);

	return (int)iocall(fid->io, _ioseek, fid->fd, n, off);
}

int
winwriteevent(Win *w, Event *e)
{
	char buf[100];
	Cfd *fid = wfid(w, "event");

	snprint(buf, sizeof buf, "%c%c%d %d \n", e->c1, e->c2, e->q0, e->q1);
	return iowrite(fid->io, fid->fd, buf, strlen(buf));
}

int
windel(Win *w, int sure)
{
	return winctl(w, sure ? "delete" : "del");
}

int
winfd(Win *w, Ioproc *io, char *name, int mode)
{
	char buf[100];

	snprint(buf, sizeof buf, "/mnt/acme/%d/%s", w->id, name);
	return ioopen(io, buf, mode);
}

static void
error(Win *w, char *msg)
{
	if(msg == nil)
		longjmp(w->jmp, 1);
	fprint(2, "%s: win%d: %s\n", argv0, w->id, msg);
	longjmp(w->jmp, 2);
}

static int
getec(Win *w, Cfd *fid)
{
	if(w->nbuf <= 0){
		w->nbuf = ioread(fid->io, fid->fd, w->buf, sizeof w->buf);
		if(w->nbuf <= 0)
			error(w, nil);
		w->bufp = w->buf;
	}
	--w->nbuf;
	return *w->bufp++;
}

static int
geten(Win *w, Cfd *fid)
{
	int n, c;

	n = 0;
	while('0'<=(c=getec(w, fid)) && c<='9')
		n = n*10+(c-'0');
	if(c != ' ')
		error(w, "event number syntax");
	return n;
}

static int
geter(Win *w, Cfd *fid, char *buf, int *nb)
{
	Rune r;
	int n;

	r = getec(w, fid);
	buf[0] = r;
	n = 1;
	if(r < Runeself)
		goto Return;
	while(!fullrune(buf, n))
		buf[n++] = getec(w, fid);
	chartorune(&r, buf);
    Return:
	*nb = n;
	return r;
}

static void
gete(Win *w, Cfd *fid, Event *e)
{
	int i, nb;

	e->c1 = getec(w, fid);
	e->c2 = getec(w, fid);
	e->q0 = geten(w, fid);
	e->q1 = geten(w, fid);
	e->flag = geten(w, fid);
	e->nr = geten(w, fid);
	if(e->nr > EVENTSIZE)
		error(w, "event string too long");
	e->nb = 0;
	for(i=0; i<e->nr; i++){
		/* e->r[i] = */ geter(w, fid, e->text+e->nb, &nb);
		e->nb += nb;
	}
/* 	e->r[e->nr] = 0; */
	e->text[e->nb] = 0;
	if(getec(w, fid) != '\n')
		error(w, "event syntax 2");
}

int
winreadevent(Win *w, Event *e)
{
	Cfd *efd;
	int r;

	if((r = setjmp(w->jmp)) != 0){
		if(r == 1)
			return 0;
		return -1;
	}
	efd = wfid(w, "event");
	gete(w, efd, e);

	/* expansion */
	if(e->flag&2){
		gete(w, efd, &w->e2);
		if(e->q0==e->q1){
			w->e2.flag = e->flag;
			*e = w->e2;
		}
	}

	/* chorded argument */
	if(e->flag&8){
		gete(w, efd, &w->e3);	/* arg */
		gete(w, efd, &w->e4);	/* location */
		strcpy(e->arg, w->e3.text);
		strcpy(e->loc, w->e4.text);
	}

	return 1;
}

int
eventfmt(Fmt *fmt)
{
	Event *e;

	e = va_arg(fmt->args, Event*);
	return fmtprint(fmt, "%c%c %d %d %d %d %q", e->c1, e->c2, e->q0, e->q1, e->flag, e->nr, e->text);
}

void*
emalloc(uint n)
{
	void *v;

	v = mallocz(n, 1);
	if(v == nil)
		sysfatal("out of memory");
	return v;
}

void*
erealloc(void *v, uint n)
{
	v = realloc(v, n);
	if(v == nil)
		sysfatal("out of memory");
	return v;
}

char*
estrdup(char *s)
{
	s = strdup(s);
	if(s == nil)
		sysfatal("out of memory");
	return s;
}

char*
evsmprint(char *s, va_list v)
{
	s = vsmprint(s, v);
	if(s == nil)
		sysfatal("out of memory");
	return s;
}

char*
esmprint(char *s, ...)
{
	char *t;
	va_list arg;
	
	va_start(arg, s);
	t = evsmprint(s, arg);
	va_end(arg);
	return t;
}

/* quick'n'dirt implementaiton */
int
threadspawnl(int *fd, char *prog, ...)
{
#define NARG	10
	va_list arg;
	char *av[NARG];
	int pid, i;

	va_start(arg, prog);
	for(i = 0; i < NARG; i++) {
		av[i] = va_arg(arg, char*);
		if(!av[i])
			break;
	}
	va_end(arg);
	pid = fork();
	if(!pid) {
		dup(fd[0], 0);
		dup(fd[1], 1);
		dup(fd[2], 2);
		exec(prog, av);
		exits("exec failed");
	}
	return pid;
}

int
pipewinto(Win *w, char *name, int errto, char *cmd, ...)
{
	va_list arg;
	char *p;
	int fd[3], pid;
	Ioproc *io;

	va_start(arg, cmd);
	p = evsmprint(cmd, arg);
	va_end(arg);
	io = ioproc();
	fd[0] = winfd(w, io, name, OREAD);
	fd[1] = dup(errto, -1);
	fd[2] = dup(errto, -1);
	pid = threadspawnl(fd, "rc", "rc", "-c", p, 0);
	free(p);
	ioclose(io, fd[0]);
	closeioproc(io);
	return pid;
}

int
pipetowin(Win *w, char *name, int errto, char *cmd, ...)
{
	va_list arg;
	char *p;
	int fd[3], pid;
	Ioproc *io;

	va_start(arg, cmd);
	p = evsmprint(cmd, arg);
	va_end(arg);
	io = ioproc();
	fd[0] = open("/dev/null", OREAD);
	fd[1] = winfd(w, io, name, OWRITE);
	if(errto == 0)
		fd[2] = dup(fd[1], -1);
	else
		fd[2] = dup(errto, -1);
	pid = threadspawnl(fd, "rc", "rc", "-c", p, 0);
	free(p);
	ioclose(io, fd[0]);
	closeioproc(io);
	return pid;
}

char*
sysrun(char *fmt, ...)
{
	static char buf[1024];
	char *cmd;
	va_list arg;
	int n, fd[3], p[2], tot;

#undef pipe
	if(pipe(p) < 0)
		sysfatal("pipe: %r");
	fd[0] = open("/dev/null", OREAD);
	fd[1] = p[1];
	fd[2] = dup(p[1], -1);

	va_start(arg, fmt);
	cmd = evsmprint(fmt, arg);
	va_end(arg);
	threadspawnl(fd, "rc", "rc", "-c", cmd, 0);

	tot = 0;
	while((n = read(p[0], buf+tot, sizeof buf-tot)) > 0)
		tot += n;
	close(p[0]);
	if(n < 0)
		return nil;
	free(cmd);
	if(tot == sizeof buf)
		tot--;
	buf[tot] = 0;
	while(tot > 0 && isspace(buf[tot-1]))
		tot--;
	buf[tot] = 0;
	if(tot == 0){
		werrstr("no output");
		return nil;
	}
	return buf;
}

static void
eventreader(void *v)
{
	Event e[2];
	Win *w;
	int i;
	
	w = v;
	i = 0;
	for(;;){
		if(winreadevent(w, &e[i]) <= 0)
			break;
		sendp(w->c, &e[i]);
		i = 1-i;	/* toggle */
	}
	sendp(w->c, nil);
	exits(nil);
}

Channel*
wineventchan(Win *w)
{
	if(w->c == nil){
		w->c = chancreate(sizeof(Event*), 0);
		threadcreate(eventreader, w, 32*1024);
	}
	return w->c;
}

Win*
newwin(void)
{
	Win *w;
	int fid;
	char buf[100];
	int id, n;

	if(io == nil){
		io = ioproc();
		if(io == nil)
			sysfatal("cannot create ioproc: %r");
	}

	fid = ioopen(io, "/mnt/acme/new/ctl", ORDWR);
	if(fid < 0)
		sysfatal("open /mnt/came/new/ctl: %r");
	n = ioread(io, fid, buf, sizeof buf);
	if(n <= 0)
		sysfatal("read /mnt/acme/new/ctl: %r");
	buf[n] = 0;
	id = atoi(buf);
	if(id == 0)
		sysfatal("read new/ctl: malformed message: %s", buf);

	w = emalloc(sizeof *w);
	w->id = id;
	wfid(w, "ctl");	/* will set w->ctl */
	w->next = nil;
	w->prev = last;
	if(last)
		last->next = w;
	else
		windows = w;
	last = w;
	return w;
}
